@@ -4,7 +4,7 @@ from pony.py23compat import itervalues, basestring
 from operator import attrgetter
 
 from pony.orm import core
-from pony.orm.core import log_sql, DBSchemaError
+from pony.orm.core import log_sql, DBSchemaError, MappingError
 from pony.utils import throw
 
 class DBSchema(object):
@@ -26,12 +26,13 @@ class DBSchema(object):
         if schema.uppercase: return s.upper().replace('%S', '%s') \
             .replace(')S', ')s').replace('%R', '%r').replace(')R', ')r')
         else: return s.lower()
-    def add_table(schema, table_name):
-        return schema.table_class(table_name, schema)
+    def add_table(schema, table_name, entity=None):
+        return schema.table_class(table_name, schema, entity)
     def order_tables_to_create(schema):
         tables = []
         created_tables = set()
-        tables_to_create = sorted(itervalues(schema.tables), key=lambda table: table.name)
+        split = schema.provider.split_table_name
+        tables_to_create = sorted(itervalues(schema.tables), key=lambda table: split(table.name))
         while tables_to_create:
             for table in tables_to_create:
                 if table.parent_tables.issubset(created_tables):
@@ -63,29 +64,28 @@ class DBSchema(object):
                                          'Try to delete %s %s first.' % (tn1, n1, tn2, n2, n2, tn2))
     def check_tables(schema, provider, connection):
         cursor = connection.cursor()
-        for table in sorted(itervalues(schema.tables), key=lambda table: table.name):
-            if isinstance(table.name, tuple): alias = table.name[-1]
-            elif isinstance(table.name, basestring): alias = table.name
-            else: assert False  # pragma: no cover
+        split = provider.split_table_name
+        for table in sorted(itervalues(schema.tables), key=lambda table: split(table.name)):
+            alias = provider.base_name(table.name)
             sql_ast = [ 'SELECT',
                         [ 'ALL', ] + [ [ 'COLUMN', alias, column.name ] for column in table.column_list ],
                         [ 'FROM', [ alias, 'TABLE', table.name ] ],
                         [ 'WHERE', [ 'EQ', [ 'VALUE', 0 ], [ 'VALUE', 1 ] ] ]
                       ]
             sql, adapter = provider.ast2sql(sql_ast)
-            if core.debug: log_sql(sql)
+            if core.local.debug: log_sql(sql)
             provider.execute(cursor, sql)
 
 class DBObject(object):
     def create(table, provider, connection):
         sql = table.get_create_command()
-        if core.debug: log_sql(sql)
+        if core.local.debug: log_sql(sql)
         cursor = connection.cursor()
         provider.execute(cursor, sql)
 
 class Table(DBObject):
     typename = 'Table'
-    def __init__(table, name, schema):
+    def __init__(table, name, schema, entity=None):
         if name in schema.tables:
             throw(DBSchemaError, "Table %r already exists in database schema" % name)
         if name in schema.names:
@@ -102,12 +102,24 @@ class Table(DBObject):
         table.parent_tables = set()
         table.child_tables = set()
         table.entities = set()
+        table.options = {}
+        if entity is not None:
+            table.entities.add(entity)
+            table.options = entity._table_options_
         table.m2m = set()
     def __repr__(table):
         table_name = table.name
         if isinstance(table_name, tuple):
             table_name = '.'.join(table_name)
         return '<Table(%s)>' % table_name
+    def add_entity(table, entity):
+        for e in table.entities:
+            if e._root_ is not entity._root_:
+                throw(MappingError, "Entities %s and %s cannot be mapped to table %s "
+                                   "because they don't belong to the same hierarchy"
+                                   % (e, entity, table.name))
+        assert '_table_options_' not in entity.__dict__
+        table.entities.add(entity)
     def exists(table, provider, connection, case_sensitive=True):
         return provider.table_exists(connection, table.name, case_sensitive)
     def get_create_command(table):
@@ -134,7 +146,16 @@ class Table(DBObject):
                 cmd.append(schema.indent+foreign_key.get_sql() + ',')
         cmd[-1] = cmd[-1][:-1]
         cmd.append(')')
+        for name, value in sorted(table.options.items()):
+            option = table.format_option(name, value)
+            if option: cmd.append(option)
         return '\n'.join(cmd)
+    def format_option(table, name, value):
+        if value is True:
+            return name
+        if value is False:
+            return None
+        return '%s %s' % (name, value)
     def get_objects_to_create(table, created_tables=None):
         if created_tables is None: created_tables = set()
         created_tables.add(table)
@@ -257,9 +278,9 @@ class DBIndex(Constraint):
             throw(DBSchemaError, 'Index %s cannot be created, name is already in use' % name)
         Constraint.__init__(index, name, schema)
         for column in columns:
-            column.is_pk = len(columns) == 1 and is_pk
-            column.is_pk_part = bool(is_pk)
-            column.is_unique = is_unique and len(columns) == 1
+            column.is_pk = column.is_pk or (len(columns) == 1 and is_pk)
+            column.is_pk_part = column.is_pk_part or bool(is_pk)
+            column.is_unique = column.is_unique or (is_unique and len(columns) == 1)
         table.indexes[columns] = index
         index.table = table
         index.columns = columns
